//=============================================================================
//	MuseScore - Embellish plugin
//
//	This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License version 2
//  as published by the Free Software Foundation and appearing in
//  the file LICENCE.GPL
//
// Preparation: Add quarter note rests in voice2 throughout. 
//				This creates segments needed by the plugin if there is no note in voice 1
//				Color the exiting chord symbols red (or any color); the added chord symbols are black
//				Adjust mingap = 480 if desired
//				Embellishments are only added if space between existing symbols is > quarter note
//				Experiment with inversions for best voice leading
//
// Known issues:
// Remove all Staff Text: they hang the plugin & MS
// 
// Limitations:
// Naked chords Bb etc are recognized as Maj; others: see embel table, adaptable to your symbols
//=============================================================================

import QtQuick 2.0
import MuseScore 3.0

MuseScore {
	version:	"0.2"
	description: "This plugin embellishes jazz lead sheet harmony"
	menuPath: "Plugins.Embellish"
	

	// create and return a new Note element with given pitch, tpc
	function createNote(pitch, tpc){
	    //console.log("pitch = ", pitch, "tpc = ", tpc)
        var note = newElement(Element.NOTE);
        note.pitch = pitch;
        note.tpc = tpc;
        note.tpc1 = tpc;
        note.tpc2 = tpc;		
        return note;
    }
	function setCursorToTime(cursor, time){
		cursor.rewind(0);
		while (cursor.segment) { 
			var current_time = cursor.tick;
			if(current_time>=time){
				return true;
			}
			cursor.next();
		}
		cursor.rewind(0);
		return false;
	}

function findindex(thishead, heads){
	    for(var i=0; i<heads.length; i++){
			if(heads[i][0] == thishead){
				return heads[i][1];
			}
		}		
        
    }
	function findhead(index, heads){
	    for(var i=0; i<heads.length; i++){
			if(heads[i][1] == index){
				return heads[i][0];
			}
		}		
        
    }
	function findheadlen(harmony){
		var str = harmony;
		var t = str.substr(1,1);
		if ((t == "b") || (t == "#")){
			return 2;
		}
		else{
			return 1;
		}
	}
	function findnewdistance(d1, d2){
		d1 = parseInt(d1);
		d2 = parseInt(d2);
		d1 = d1 + d2;
		//console.log("d1 = ", d1);
		if (d1 > 12) d1 -= 12;
		if (d1 < 1) d1 += 12;
		return d1;
	}

	onRun: {
		var mingap = 480;
		var cursor = curScore.newCursor()
		if (typeof curScore === 'undefined') {
			console.log("Please open a jazz lead sheet before calling execution of this function.");
			Qt.quit();
		}
		cursor.voice = 1;
		
		console.log("Generating jazz harmony embellishments");

		//0: chord pair:"chordtype1 distance chordtype2"
		//1: distance to embed chord in half notes: from -5 (eg C-G) to 6 (eg C- F#)
		//2: embed chord type
		//3-4 etc repeats 1-2 ie other possible embed chords
var embel = [
["7 0 7b9",-5,"m7"],
["7 0 7sus",-5,"m7"],
["7 0 dim",-1,"7"],
["7 0 m7",0,"7b13",1,"7",1,"dim",-5,"m7b5",6,"7"],
["7 0 m7b5",0,"7b13",1,"dim",-5,"m7b5",6,"7"],
["7 -1 7",1,"7",-1,"7sus",5,"7",6,"7"],
["7 1 7",-4,"7",5,"7",-5,"m7"],
["7 1 7#11",3,"7",-5,"m7"],
["7 -1 7b9",6,"m7b5",6,"m7"],
["7 -1 7sus",6,"m7"],
["7 1 dim",0,"m7"],
["7 1 m7",-1,"7",3,"7",4,"m7"],
["7 -1 m7",3,"7",5,"7",6,"7"],
["7 -1 m7b5",-1,"7",-2,"7",-3,"dim",5,"7",-4,"7#11"],
["7 1 m7b5",4,"m7",5,"7"],
["7 -1 Maj",-2,"7",6,"m7b5",-3,"7",-3,"7#11",4,"7",5,"7",6,"7"],
["7 1 Maj",-4,"7",5,"7",2,"7"],
["7 -2 7",-1,"7",1,"7",-1,"m7b5",-4,"7",5,"7",6,"7"],
["7 2 7",1,"7",-1,"m7b5",3,"7",-3,"m7"],
["7 2 7b9",-3,"m7b5"],
["7 -2 dim",-3,"m7"],
["7 2 m7",-1,"7#11",1,"dim",-2,"7",-2,"mMaj7",3,"7",-3,"7",5,"7"],
["7 -2 m7",-3,"dim",-4,"7",5,"7"],
["7 2 m7b5",1,"7",6,"7"],
["7 -2 Maj",-1,"7",1,"dim",-3,"7",3,"mMaj7"],
["7 2 Maj",3,"7",1,"7"],
["7 3 7",-2,"7",-2,"m7",-3,"7",5,"7",-5,"m7",6,"7"],
["7 -3 7",-2,"7",3,"7",-4,"7",4,"m7b5",5,"7",6,"7"],
["7 3 7#11",4,"m7b5"],
["7 3 m7",1,"7",-1,"7",-2,"7#11",5,"7"],
["7 -3 m7",-2,"7",-2,"m7",3,"7",6,"7"],
["7 3 m7b5",6,"7"],
["7 -3 m7b5",-2,"7",-2,"m7b5",-3,"m7",6,"7"],
["7 3 Maj",1,"7",-1,"7"],
["7 3 mMaj7",0,"m7"],
["7 -4 7",0,"m7",1,"7",-2,"7",3,"7",-3,"m7",5,"7"],
["7 4 7",-1,"7",2,"7",-3,"7",5,"7",-5,"7"],
["7 4 7sus",-1,"m7b5",5,"7"],
["7 4 m7",-1,"7",-2,"7",5,"7",6,"7",6,"7#11"],
["7 -4 m7",2,"7",-5,"7",5,"7",6,"7"],
["7 -4 m7b5",-1,"7"],
["7 4 m7b5",-1,"7b9",3,"7",5,"7",-5,"m7"],
["7 4 Maj",-1,"7",-1,"7b13",-2,"7",3,"7",-3,"m7b5",5,"7"],
["7 -4 Maj",1,"dim",3,"7",5,"7",6,"7"],
["7 4 mMaj7",-1,"7"],
["7 5 69",6,"7"],
["7 -5 7",-2,"7",2,"7",3,"7",-4,"7",5,"7",6,"7"],
["7 5 7#11",-1,"7"],
["7 5 7sus",0,"m7",1,"dim",6,"7"],
["7 5 dim",-1,"7b9",6,"7"],
["7 -5 m7",-1,"7"],
["7 -5 m7b5",1,"7"],
["7 5 m7b5",-1,"7",6,"7"],
["7 -5 Maj",1,"dim",2,"7",-4,"7",5,"7"],
["7 5 mMaj7",1,"7"],
["7 6 7#11",-5,"m7"],
["7 6 7b9",-1,"7"],
["7 6 dim",-5,"m7"],
["7 6 m7",5,"7"],
["7 6 m7b5",-1,"7",3,"7",5,"7"],
["7 6 Maj",1,"7",-1,"7"],
["7#11 0 7",2,"7"],
["7#11 -1 7",-1,"7sus",5,"7"],
["7#11 2 Maj",1,"dim",-2,"m7b5",3,"7",5,"7"],
["7#11 3 m7",5,"7"],
["7#11 -3 m7",-1,"7"],
["7#11 4 7",-1,"7",5,"7"],
["7b9 0 m7b5",-3,"m7"],
["7b9 4 m7b5",5,"7"],
["7b9 5 m7",0,"7",6,"7"],
["7sus 0 7",0,"dim"],
["7sus 0 dim",1,"m7",-5,"m7"],
["7sus -1 m7",2,"m7",5,"Maj"],
["7sus -3 m7",0,"7b9"],
["dim 1 7",1,"7sus",4,"7"],
["dim 1 m7",2,"7"],
["dim -3 Maj",4,"7"],
["m7 0 7",1,"7",5,"7"],
["m7 0 m7b5",3,"m7",3,"mMaj7",6,"7"],
["m7 -1 7",5,"7",5,"7b9",6,"7",6,"7#11"],
["m7 1 7",2,"m7",2,"m7b5",-4,"7",-4,"mMaj7",5,"7"],
["m7 -1 7#11",5,"7",6,"7"],
["m7 -1 7b9",5,"m7"],
["m7 -1 dim",-1,"7",5,"7",5,"7b9"],
["m7 1 dim",5,"7"],
["m7 1 m7",-1,"m7",-4,"7",5,"7"],
["m7 -1 m7",5,"7"],
["m7 -1 Maj",-1,"7",1,"7",-4,"7",5,"7",5,"m7"],
["m7 1 Maj",-1,"7",-4,"7",5,"7",5,"m7"],
["m7 -2 7",-1,"7",4,"dim",5,"7"],
["m7 2 7",-3,"m7b5",-4,"7",5,"7"],
["m7 2 7b9",-3,"m7b5"],
["m7 -2 7sus",-1,"7"],
["m7 -2 m7",0,"dim",-1,"7",-1,"m7",4,"7",5,"7"],
["m7 2 m7",0,"m7b5",1,"dim",3,"mMaj7",5,"7"],
["m7 2 m7b5",0,"m7b5",-1,"7",3,"7",3,"7#11",5,"7",-3,"m7b5",1,"7",-4,"7"],
["m7 -2 Maj",0,"m7b5",-2,"dim",3,"mMaj7",5,"7"],
["m7 3 7",2,"m7",5,"7",6,"7"],
["m7 -3 7",4,"m7b5",5,"7"],
["m7 3 7#11",-1,"7",5,"7"],
["m7 -3 7b9",4,"m7b5"],
["m7 3 m7",0,"m7b5",0,"mMaj7",1,"7#11/10",2,"m7",3,"m7b5",3,"mMaj7",5,"7"],
["m7 -3 m7",5,"7"],
["m7 -3 m7b5",5,"7",0,"m7b5",0,"mMaj7",6,"7"],
["m7 3 Maj",1,"7"],
["m7 3 mMaj7",2,"m7",3,"Maj"],
["m7 -4 7",-1,"7",1,"7",-1,"dim",3,"7",-3,"m7b5",3,"mMaj7",5,"7"],
["m7 -4 7#11",-3,"m7b5",5,"7"],
["m7 4 dim",5,"7"],
["m7 4 m7",5,"7"],
["m7 -4 m7",-1,"7",5,"7"],
["m7 4 m7b5",1,"7",5,"7"],
["m7 -4 Maj",-3,"m7b5"],
["m7 -4 mMaj7",-3,"m7b5"],
["m7 -5 7",1,"7",2,"m7b5",-4,"7",5,"7"],
//["m7 5 7",6,"7"],
["m7 5 7#11",6,"7",6,"7#11"],
["m7 5 7b13",6,"7"],
["m7 -5 7b9",2,"m7"],
["m7 5 7sus",6,"7"],
["m7 5 m7",0,"7",0,"mMaj7",5,"7",6,"7",-1,"7"],
["m7 -5 m7",0,"mMaj7",2,"7",-4,"7",5,"7",5,"7sus"],
["m7 -5 m7b5",1,"7",5,"7"],
["m7 5 m7b5",5,"7"],
["m7 -5 Maj",5,"7"],
["m7 6 7",1,"7",5,"7"],
["m7 6 m7",5,"7"],
["m7b5 -1 7",1,"7",5,"7b9"],
["m7b5 -1 m7",-1,"mMaj7"],
["m7b5 -2 dim",5,"7"],
["m7b5 -2 m7",-1,"7",-2,"m7b5/6"],
["m7b5 -2 Maj",-1,"7",-3,"dim",5,"7",5,"7sus"],
["m7b5 3 m7",-1,"Maj"],
["m7b5 -3 Maj",-1,"7"],
["m7b5 -4 m7",-1,"7"],
["m7b5 5 7",0,"7",0,"m7",-1,"7",5,"7sus",6,"7"],
["m7b5 5 7b9",5,"7",6,"7"],
["m7b5 5 7sus",-1,"7"],
["m7b5 5 m7",-1,"7",6,"7"],
["m7b5 5 m7b5",-1,"7",6,"7"],
["Maj 0 7",-1,"7",1,"m7b5",4,"m7",-5,"m7b5",-5,"m7",6,"7"],
["Maj 0 7sus",0,"7"],
["Maj 0 dim",0,"m7"],
["Maj 0 m7",0,"dim",0,"mMaj7",-2,"dim"],
["Maj 0 mMaj7",-3,"m7"],
["Maj -1 7",0,"7",0,"m7b5",-1,"7sus",6,"m7b5"],
["Maj 1 7",2,"m7",-3,"m7",-4,"7"],
["Maj 1 dim",0,"mMaj7",-2,"7",-3,"m7",4,"m7"],
["Maj -1 m7",0,"7",0,"m7",0,"mMaj7",-2,"7",2,"7",-2,"7sus",0,"7sus",2,"7sus",5,"7",5,"7sus/5",4,"7/7"],
["Maj 1 m7",0,"7",4,"7",0,"7#11",5,"7"],
["Maj -1 m7b5",0,"7",5,"7",5,"Maj"],
["Maj 1 m7b5",-2,"7",-3,"m7",5,"7"],
["Maj 1 Maj",-4,"7"],
["Maj -1 Maj",5,"7",6,"7"],
["Maj -2 7",-1,"7",-1,"m7b5",-1,"m7",-3,"m7",4,"7",5,"7"],
["Maj 2 7",-3,"m7",5,"7"],
["Maj -2 7#11",-3,"m7"],
["Maj 2 7#11",-3,"m7"],
["Maj 2 7b9",-3,"m7"],
["Maj 2 7sus",1,"m7b5",-3,"m7"],
["Maj -2 dim",0,"mMaj7"],
["Maj -2 m7",0,"7",0,"dim",6,"7"],
["Maj 2 m7",-5,"m7b5/6",0,"dim",1,"dim",3,"7",-4,"7"],
["Maj 2 m7b5",3,"7",-3,"m7",5,"7"],
["Maj 3 7",0,"7",-2,"7",-2,"7#11",-3,"7",-3,"m7",4,"7",4,"m7b5",4,"m7"],
["Maj -3 7",0,"7",-2,"7",4,"m7",5,"7"],
["Maj -3 7b9",-2,"7",4,"m7b5",4,"m7"],
["Maj 3 dim",4,"m7"],
["Maj -3 m7",-2,"7",5,"7",5,"7#11",5,"7sus",3,"7",4,"7",-4,"dim"],
["Maj 3 m7",2,"m7",-3,"m7",4,"m7"],
["Maj -3 m7b5",0,"m7",-3,"m7"],
["Maj -3 Maj",-2,"7#11"],
["Maj 3 Maj",-2,"dim"],
["Maj -3 mMaj7",5,"7"],
["Maj -4 7",0,"7sus",-2,"7",3,"7",-3,"m7"],
["Maj 4 7",-1,"m7b5",-2,"7",5,"7"],
["Maj 4 7#11",5,"7"],
["Maj 4 7b9",-1,"m7b5"],
["Maj 4 m7",-2,"7",2,"m7",-3,"m7",5,"7",-5,"7sus"],
["Maj -4 m7",-3,"m7"],
["Maj 4 m7b5",-2,"7",-2,"7#11",2,"7b9",2,"m7",4,"m7",5,"7",5,"7#11",-5,"m7b5"],
["Maj -4 Maj",3,"7",-3,"m7"],
["Maj 4 Maj",5,"7"],
["Maj 5 7",0,"7",0,"mMaj7",6,"m7b5"],
["Maj -5 7",2,"m7",-4,"7",6,"dim",1,"7"],
["Maj 5 7#11",0,"mMaj7",6,"m7b5"],
["Maj -5 7sus",-3,"m7"],
["Maj 5 m7",0,"7",0,"7b9",2,"m7",-3,"m7",4,"m7",6,"7",6,"m7b5"],
["Maj -5 m7",1,"7",2,"m7",-3,"m7",-4,"7",6,"7",6,"dim"],
["Maj -5 m7b5",-2,"7"],
["Maj 5 Maj",-2,"7",0,"7",0,"7#11",-2,"7#11",6,"7",4,"7#11"],
["Maj -5 Maj",0,"mMaj7",1,"dim",2,"7",5,"7"],
["Maj 5 mMaj7",5,"Maj",6,"m7b5"],
["Maj 6 7",0,"7",1,"7#11/10",1,"m7b5",5,"7",-5,"m7"],
["Maj 6 7#11",1,"m7b5"],
["Maj 6 7b9",1,"m7b5"],
["Maj 6 dim",5,"Maj"],
["Maj 6 m7",0,"7",-3,"7",-3,"m7",5,"7"],
["Maj 6 m7b5",-2,"7",-3,"7",-3,"m7",5,"7",-5,"7",-5,"m7b5"],
["mMaj7 -1 7",0,"m7",5,"7"],
["mMaj7 -1 m7",0,"m7",5,"7"],
["mMaj7 2 m7",-3,"m7b5"],
["mMaj7 2 m7b5",-3,"m7b5"],
["mMaj7 3 7",5,"7"],
["mMaj7 -3 7",5,"7"],
["mMaj7 -3 m7b5",0,"m7"],
["mMaj7 -4 7",0,"m7",-3,"m7b5"],
["mMaj7 4 m7b5",1,"7"],
["mMaj7 5 7",0,"m7"],
["mMaj7 -5 Maj",-2,"dim",5,"7",2,"7"],
];
		
var heads = [
["C",1],
["D",3],
["E",5],
["F",6],
["G",8],
["A",10],
["B",12],
["Db",2],
["Eb",4],
["Fb",5],
["Gb",7],
["Ab",9],
["Bb",11],
["Cb",12],
["B#",1],
["C#",2],
["D#",4],
["E#",6],
["F#",7],
["G#",9],
["A#",11],
];
		// collect the existing chord symbols and their segment.tick
		var chordnames = new Array(128); 
		// chordnames
		// O: chordname
		// 1: tick
		// 2: head
		// 3: tail
		for(var i=0; i<128; i++) chordnames[i]=new Array();
		var count = 0;
		for (var track = 0; track < /*curScore.ntracks*/ 1; ++track) {
			var segment = curScore.firstSegment();
			while (segment) {
				var aCount = 0;
				var annotation = segment.annotations[aCount];
		
				while (annotation) {
					if (annotation.type == Element.HARMONY) {
						var harmony =  annotation;
						var str = harmony.text;
						var n = str.search("/");
							if (n > -1){
								console.log('str = ' + str);
								str = str.substr(0,n);
								console.log('str = ' + str);
							}
						chordnames[count][0] = str;
						chordnames[count][1] = segment.tick;
						var headlen = findheadlen(str);
						var hstr = str.substr(0,headlen);
						// make the root uppercase
						hstr = hstr.toUpperCase();
						if (hstr.substr(1,1) == "B"){
							hstr = hstr.substr(0,1) + "b";
						}
						chordnames[count][2] = hstr;
						if (str.length - headlen > 0) chordnames[count][3] = str.substr(headlen,(str.length - headlen))
						else chordnames[count][3] = "Maj";
						// change alt to 7
						var n = chordnames[count][0].search("7alt");
							if (n > -1){
								var i1 = findindex(chordnames[count][2], heads);
								var i2 = findnewdistance(i1, 6);
								chordnames[count][2] = findhead(i2, heads)
								chordnames[count][3] = "7";
							}
						console.log("head -", chordnames[count][2],"-");
						console.log("tail -",str.substr(headlen,(str.length - headlen)),"-");
						console.log('harmony.text = ' + harmony.text);
						console.log('time = ' + segment.tick);
						annotation = segment.annotations[++aCount];
						count++;					
					}
				}
				segment = segment.next;
			}
		}
		//add a whole note of time to the last measure
		chordnames[count][1] = chordnames[count-1][1] + 1920;
		
		// construct the chord pairs
		for (var c = 0; c < count-1; c++){
			var gap = chordnames[c+1][1] - chordnames[c][1];
			if (gap > mingap){
				var i1 = findindex(chordnames[c][2], heads);
				var i2 = findindex(chordnames[c+1][2],heads);
				var distance = i2 - i1;
				if (distance < -5) distance += 12;
				if (distance > 6) distance -= 12;
				
				// q = "tail1 distance tail2"
				var q = chordnames[c][3] + " " + distance + " " + chordnames[c+1][3];
				console.log("q =", q);
				// search for q in the embel table
				for (var d = 0; d < embel.length; d++){
					// this stacks embellishments halfway between existing symbols
					if (embel[d][0] == q){
						for (var i = 1; i < embel[d].length; i+=2){
							var newdist = findnewdistance (i1,embel[d][i]);
							var newhead = findhead(newdist, heads);
							var newchord = newhead + embel[d][i+1];
							// is it a slash chord?
							// in the embel table the /base is notated as offset from the root in halftones
							var n = newchord.search("/");
							if (n > -1){
								var slashi = newchord.substr(n+1, newchord.length - (n+1));
								var slashi2 = findnewdistance (slashi,newdist);
								var shead = findhead(slashi2, heads);
								newchord = newchord.substr(0,n+1) + shead;
							}
							var newtime = chordnames[c][1] + (gap/2);
							setCursorToTime(cursor, newtime); 
							
							var harmony = newElement(Element.HARMONY);
            				harmony.text = newchord;
							console.log("newchord -", newchord);
            				cursor.add(harmony);
						}
					}
				}
			}
			
		}
		Qt.quit();
	}// end on Run
}
